R<x>:=PolynomialRing(Rationals());

function FindMinimalPolynomial(number, bound);
	for i in [1..bound] do
		PowerBasis:=[number^k : k in [0..i]];
		try
			coeff:=IntegerRelation(PowerBasis, 10^5);
			if Abs(&+[coeff[k]*number^(k-1) : k in [1..i+1]]) le 10^(-100) then
				return coeff;
			end if;
		catch e;
		end try;
	end for;
	return [];
end function;

function MatrixRatio(M1,M2);
    for i in [1..4] do
        for j in [1..4] do
            if M2[i][j] ne 0 then
                return M1[i][j]/M2[i][j];
            end if;
        end for;
    end for;
end function;


function RecognizeNumberField(M);
	for i in [1..2] do
		for j in [1..2] do
			e:=M[i][j];
			coeff:=FindMinimalPolynomial(e,6);
			poly := PolynomialRing(Integers())!coeff;
			if IsIrreducible(poly) and Degree(poly) ge 2 then
				// "Polynomial defining the number field:", poly;
				K<a>:=NumberField(poly);
				return K, a;
			end if;
		end for;
	end for;
	return Rationals(), 1;
end function;

function IsProjection(M);
	if M eq ZeroMatrix(Integers(),4,4) then
		return false;
	end if;
	ratio:=MatrixRatio(M^2,M);
	return M^2 eq ratio*M;
end function;

function FindProjector(endo, alpha);
	for counter in [1..#Generators(endo)] do
		m:=endo.counter;
		a:=alpha[counter];
		ev:=Eigenvalues(m);
		if not IsEmpty(ev) then
			eigVal:={@ z : z in ev@};
			TargetEigenvalue:=eigVal[1][1];
	
			MMatrix := m - TargetEigenvalue;
			AMatrix := a - TargetEigenvalue;
			if IsProjection(MMatrix) then
				MapDegree:= Integers()!Abs(MatrixRatio(MMatrix^2,MMatrix)) ;
				return true, MMatrix, AMatrix, MapDegree;
			end if;
		end if;
	end for;






	if #Generators(endo) eq 2 then
		for bound in [1..4] do
		for a in [-bound..bound] do
			for b in [-bound..bound] do
			for c in [-bound..bound] do
				m:=endo.1*a+endo.2*b+(endo.1*endo.2)*c;
				alphaMatrix:=alpha[1]*a+alpha[2]*b+(alpha[1]*alpha[2])*c;
				ev:=Eigenvalues(m);
				if not IsEmpty(ev) then
					eigVal:={@ z : z in ev@};
					TargetEigenvalue:=eigVal[1][1];
	
					MMatrix := m - TargetEigenvalue;
					AMatrix := alphaMatrix - TargetEigenvalue;
					if IsProjection(MMatrix) then
						MapDegree:= Integers()!Abs(MatrixRatio(MMatrix^2,MMatrix)) ;
						if MapDegree ne 0 then
							return true, MMatrix, AMatrix, MapDegree;
						end if;
					end if;
				end if;
			end for;
			end for;
		end for;
		end for;
	end if;


	if #Generators(endo) eq 3 then
		endo.1;
		endo.2;
		endo.3;
		for bound in [1..3] do
		for a in [-bound..bound] do
			for b in [-bound..bound] do
			for c in [-bound..bound] do
				m:=endo.2*a+endo.3*b+endo.2*endo.3*c;
				alphaMatrix:=alpha[2]*a+alpha[3]*b+alpha[2]*alpha[3]*c;
				ev:=Eigenvalues(m);
				if not IsEmpty(ev) then
					eigVal:={@ z : z in ev@};
					TargetEigenvalue:=eigVal[1][1];
	
					MMatrix := m - TargetEigenvalue;
					AMatrix := alphaMatrix - TargetEigenvalue;
					if IsProjection(MMatrix) then
						MapDegree:= Integers()!Abs(MatrixRatio(MMatrix^2,MMatrix)) ;
						if MapDegree ne 0 then
							return true, MMatrix, AMatrix, MapDegree;
						end if;
					end if;
				end if;
			end for;
			end for;
		end for;
		end for;
	end if;



	return false, _, _, _;
end function;



function ReconstructEllipticCurve(ECLattice, K, a : Mode:=0);

	InternalC:=ComplexField(600);
	if Im(ECLattice[1][2]/ECLattice[1][1]) ge 0 then
		ELat:=[InternalC!ECLattice[1][1],InternalC!ECLattice[1][2]];
	else
		ELat:=[InternalC!ECLattice[1][2],InternalC!ECLattice[1][1]];
	end if;

	d:=Degree(K);

	E4:=Eisenstein(4,ELat);
	zeta := RiemannZeta( : Precision:=600);
	G4:=2*E4*Evaluate(zeta,4) / ELat[1]^4;


	E6:=Eisenstein(6,ELat);
	G6:=2*E6*Evaluate(zeta,6) / ELat[1]^6;

	// "G4", 60*G4;
	// "G6", 140*G6;

	bestFit:=InternalC!1;
	for EmbeddingIndex in [1..d] do
		EmbeddedA:=InternalC!Conjugates(a)[EmbeddingIndex];
		PowerBasis2:=[G4] cat [EmbeddedA^k : k in [0..d-1]];
		LinRel2:=IntegerRelation(PowerBasis2, 10^18);

		fit:=&+([G4 * LinRel2[1]] cat [LinRel2[k]*EmbeddedA^(k-2) : k in [2..d+1]]);
		if(Abs(fit) lt Abs(bestFit)) then
			BestEmbedding:=EmbeddingIndex;
			bestFit:=fit;
		end if;
	end for;

	EmbeddedA:=InternalC!Conjugates(a)[BestEmbedding];
	PowerBasis2:=[G4] cat [EmbeddedA^k : k in [0..d-1]];
	LinRel2:=IntegerRelation(PowerBasis2, 10^18);

	Coeff2:=[-LinRel2[k]/LinRel2[1] : k in [2..d+1]];

	ZP<x>:=PolynomialRing(Integers());
	
	PowerBasis3:=[G6] cat [EmbeddedA^k : k in [0..d-1]];
	LinRel3:=IntegerRelation(PowerBasis3,10^18);
	Coeff3:=[-LinRel3[k]/LinRel3[1] : k in [2..d+1]];

	

	g2:=60*K!Coeff2;
	g3:=140*K!Coeff3;

	g2Embedded:= &+ [-LinRel2[k+1]/LinRel2[1] * EmbeddedA^(k-1) : k in [1..d] ];
	g3Embedded:= &+ [-LinRel3[k+1]/LinRel3[1] * EmbeddedA^(k-1) : k in [1..d] ];

	S<x>:=PolynomialRing(K);
	f:=x^3-4*g2*x-16*g3;

	EAlg:=EllipticCurve(f);

	C300:= Parent(ECLattice[1][1]);
	P<w>:=PolynomialRing(C300);
	EAnComputed:=EllipticCurve(w^3-4*C300!(60*G4)*w-16*C300!(140*G6));
	EAn:=EllipticCurve(w^3-4*C300!(60*g2Embedded)*w-16*C300!(140*g3Embedded));


	/*
	print "=================";
	EAlg;
	EAn;
	"Difference g2 is small?", (Abs(g2Embedded-G4) );
	"Difference g3 is small?", (Abs(g3Embedded-G6) );
	"Difference g2 is small?", (Abs(g2Embedded-G4) le 10^(-50));
	"Difference g3 is small?", (Abs(g3Embedded-G6)  le 10^(-50));
	"Original j-Invariant", jInvariant(ELat[2]/ELat[1]);
	"Analytic j-Invariant", jInvariant(EAn);
	"Algebraic j-Invariant", jInvariant(EAlg);
	"Difference j-Invariants", (Abs(jInvariant(EAn)-jInvariant(ELat[2]/ELat[1])));
	print "=================";
	*/
	if Mode eq 1 then
		return EAlg, EAnComputed, EmbeddedA;
	end if;
	return EAlg, EAn, EmbeddedA;
end function;



function ExtractLatticeGenerators(M);
// Input: a 1 x n matrix with entries in C whose "columns" should span a rank-2 lattice
// Output: a 2 x n submatrix of M that generates the same lattice

g1:=0;
g2:=0;

C := BaseRing(M); // recover the complex field containing the coefficients of M
eps := 10^(-4); // numbers smaller than eps in absolute values are rounded to 0 to decide linear dependence
n := #Rows(Transpose(M));

// Loop through the coefficients of M to find the first nonzero entry
for i in [1..n] do
    if Abs(M[1][i]) ge eps then
        g1 := M[1][i];
        break i;
    end if;
end for;

if g1 eq 0 then
	return Matrix(C,1,2,[0,0]), false;
end if;

// Loop through the coefficients of M to find an entry that is not RR-linearly dependent from g1
for i in [1..n] do
    if Abs(M[1][i]) ge eps and Abs(Im(M[1][i]/g1)) ge eps then
        g2 := M[1][i];
        break i;
    end if;
end for;

// the rational span of g1, g2 should be the same as the rational span of all the entries of M.
// Now loop through the entries of M once again and express them as rational linear combinations of g1, g2

CoordinateVectors:=Matrix(Rationals(),2,n,[]);
for i in [1..n] do
        FindLinearRelation := [g1,g2,M[1][i]];
	CoefficientsLinearRelation:=LinearRelation(FindLinearRelation);
    if CoefficientsLinearRelation[3] ne 0 then
CoordinateVectors[1][i]:=-CoefficientsLinearRelation[1]/CoefficientsLinearRelation[3];
CoordinateVectors[2][i]:=-CoefficientsLinearRelation[2]/CoefficientsLinearRelation[3];
    end if;
end for;

// Let's recover the lattice itself
L:=Lattice(Transpose(CoordinateVectors));
// Find a basis of the lattice
B:=Basis(L);

// h1, h2 should now be complex numbers that generate the same lattice as all the entries of M
h1:=B[1][1]*g1+B[1][2]*g2;
h2:=B[2][1]*g1+B[2][2]*g2;

// return h1, h2 in matrix form
return Matrix(C,1,2,[h1,h2]), true;
end function;


function ComputeMap(polynomial : WhichEigenvalue:=1 );
// The sextic such that the genus 2 curve C is y^2=f(x)
f1 := polynomial;

// Print the equation of the genus 2 curve
"Genus 2 curve C: ", HyperellipticCurve(f1);

for precision in [1..5] do // to speed up thing and to help compensate approximation errors, work at different precisions
for Mode in [0..1] do // the two working modes of ReconstructEllipticCurve
try
"Working with", 100*precision, "decimal digits";
// General definition
S<x> := PolynomialRing(RationalField());
C<i> := ComplexField(100*precision);
KC<xc> := PolynomialRing(C);

// Setup the analytic Jacobian of C
f1C := Evaluate(f1,xc);
A1 := AnalyticJacobian(f1C);

// PointAtInfinity:=ToAnalyticJacobian(C!10^90,Sqrt(Evaluate(f1C,10^90)),A1);

// Use MAGMA's functionalities to compute numerically the endomorphism ring of Jac(C)
endo, alpha:=EndomorphismRing(A1);

// Define the MMatrix matrices by hand. Here if P is the period matrix we have AP=PM, so A is 2x2 (complex, or rather with coefficients in the field of definition of the morphism from Jac(C) to E) and M is 4x4 (integral)

ProjectorFound, MMatrix, AMatrix, MapDegree := FindProjector(endo, alpha);

if ProjectorFound eq false then
	error "Could not compute a projection matrix";
end if;

"Projection matrix", MMatrix;
// "AMatrix", AMatrix;
"Map degree:", MapDegree;



K,a:=RecognizeNumberField(AMatrix);
// "Number field of definition:", K;
d:=Degree(K);

// P1 is the BigPeriodMatrix of Jac(C). The columns of P2 are supposed to generate a rank-2 sublattice of the lattice defining Jac(C) in C^4; this sublattice should correspond to the elliptic curve we are projecting onto.

P1 := BigPeriodMatrix(A1);
P2 := P1*Matrix(C,MMatrix);

// Now we project the elliptic curve lattice down to C. We do this by the projection map C^2 \to C given by (x_1, x_2) \mapsto x_1. One can show that one of the three maps (x_1, x_2) \mapsto x_1, (x_1, x_2) \mapsto x_2, (x_1, x_2) \mapsto x_1 + x_2 will give a rank-2 lattice in C.

	ProjectionVector := Matrix(C,1,2,[C!1,C!1]);     // the linear projection "is" taking the product on the left with ProjectionVector
	ELatticeC2 := ProjectionVector*P2;

	// We now have four complex numbers that should generate a rank-2 lattice in C. Extract two of them that still generate the same lattice.
	ELattice, success := ExtractLatticeGenerators(ELatticeC2);

if success eq false then
	ProjectionVector := Matrix(C,1,2,[C!1,C!0]);     // the linear projection "is" taking the product on the left with ProjectionVector
	ELatticeC2 := ProjectionVector*P2;
	ELattice, success := ExtractLatticeGenerators(ELatticeC2);
end if;

if success eq false then
	ProjectionVector := Matrix(C,1,2,[C!0,C!1]);     // the linear projection "is" taking the product on the left with ProjectionVector
	ELatticeC2 := ProjectionVector*P2;
	ELattice := ExtractLatticeGenerators(ELatticeC2);
end if;


// Reconstruct a rational equation for the quotient elliptic curve E
E,EAn,EmbeddedA:=ReconstructEllipticCurve(ELattice, K, a : Mode:=Mode);
T<x>:=PolynomialRing(K);

// Compute a (simplified) Weierstrass model for E, y^2=f2(x)
f2:=HyperellipticPolynomials(WeierstrassModel(E));

// Print the equation of the EllipticCurve quotient
WeierstrassModel(E);

// Account for extra endomorphisms of E; these can get in the way
// when MAGMA computes an identification of the BigPeriodMatrix of
// A2 with the ECLattice
j:=jInvariant(E);
ListAutomorphisms:=[C!1];

if j eq 0 then
	ListAutomorphisms:= [C!1, C!(1-Sqrt(-3))/2, C!(1+Sqrt(-3))/2];
end if;
if j eq 1728 then
	ListAutomorphisms:= [C!1, i];
end if;

// Setup the AnalyticJacobian of the elliptic curve
f2C:=HyperellipticPolynomials(EAn);
A2 := AnalyticJacobian(f2C);

// f2C;

// Find an iso between the BigPeriodMatrices of ELattice (which lives in a quotient of the AnalyticJacobian of C)
// and the AnalyticJacobian of the elliptic curve E

iso2,MMatrix2,AlphaMatrix2:=IsIsomorphicBigPeriodMatrices(ELattice,BigPeriodMatrix(A2));

"Checking for isomorphism:", iso2;

if iso2 eq false then  // try swapping the two columns that generate the lattice in \mathbb{C}
    e1:=ELattice[1][1];
    e2:=ELattice[1][2];
    ELattice[1][1]:=e2;
    ELattice[1][2]:=e1;
iso2,MMatrix2,AlphaMatrix2:=IsIsomorphicBigPeriodMatrices(ELattice,BigPeriodMatrix(A2));
    "Checking again for isomorphism:", iso2;
    if(iso2 eq false) then
//      print "Could not find an isomorphism between the elliptic curve lattices";
	error "Could not find an isomorphism between the elliptic curve lattices";
    end if;
end if;


// Ok, we're all set. We can now map points from C to AnalyticJac(C) (using ToAnalyticJacobian),
// to the elliptic curve corresponding to ELattice (multiplying by ProjectionVector),
// then to the AnalyticJacobian of the elliptic curve E (applying AlphaMatrix2),
// and finally to E itself (using FromAnalyticJacobian)

// SmallC:=ComplexField(5);

for iso in ListAutomorphisms do
for TranslateX in [0..1] do
for TranslateY in [0..1] do

LinearEquationMatrix:=Matrix(K,0,2*(1+MapDegree),[]); // matrix that gives the linear system satisfied by the coefficients of the first coordinate of the map C \to E
counter := 0; // number of images computed so far

for x0 in [-MapDegree-3..MapDegree+2] do

    counter := counter + 1;
    y0:=Evaluate(f1,x0);
    // "Computing image of point (", x0, ", ", " Sqrt(", y0, ") )";

    p1:=ToAnalyticJacobian(C!x0,C!Sqrt(y0),A1);     // point on the Jacobian corresponding to the point (x0, \sqrt(y0)) on the curve
    // "p1", p1;
    // p1:=p1-PointAtInfinity;
    stage2:=AMatrix*p1;                     // image of p1 under the endomorphism of Jac(C) that projects to the elliptic curve embedded in the Jacobian
    // "Stage 2", stage2;
    stage3:=ProjectionVector*stage2;             // image of p1 in the 1-dimensional C-vector space which contains the ELattice
    // "Stage 3", stage3;
    stage4:=iso*AlphaMatrix2*stage3;                 // image of p1 in the 1-dimensional C-vector space which covers E
    // "Stage 4", stage4;

    // a basis of the 2-torsion of E
    Column1:=Matrix(C,1,1,[BigPeriodMatrix(A2)[1][1]]);
    Column2:=Matrix(C,1,1,[BigPeriodMatrix(A2)[1][2]]);

    // translate by a 2-torsion point to improve form of the map
    // p2:=FromAnalyticJacobian(stage4,A2); // this is (supposed to be) the image of the point p1 in the 1-dimensional complex vector space that uniformizes E

    try

    p2:=FromAnalyticJacobian(stage4+TranslateX/2*Column1+TranslateY/2*Column2,A2); // this is (supposed to be) the image of the point p1 in the 1-dimensional complex vector space that uniformizes E

    if(IsDefined(p2,1)) then // special case: p1 maps to the point at infinity of E
    if(Abs(p2[1][1]) le 10^50) then
        // coordinates (x,y) of the image point
        xi:=p2[1][1];
        yi:=p2[1][2];

	basis:=[xi] cat [EmbeddedA^k : k in [0..d-1]];


        // use LLL to recognize xi as an exact algebraic number in Q(a)
	// "Basis", [SmallC!b : b in basis];
	Relation:=IntegerRelation(basis,10^15);
	coeff:=[-Relation[k]/Relation[1] : k in [2..d+1]];
	xi := K!coeff;
        if (Height(Norm(xi)) le 10^(10*d)) then
            // "x-coordinate of the image point (numerically evaluated): ", xi;

            MonomialsNumerator := [ x0^i : i in [0..MapDegree] ] ;                         // so the numerator is \sum a_n x^n
            MonomialsDenominator := [ -xi * x0^i : i in [0..MapDegree] ] ;                     // and it should be equal to +xi \sum b_n x^n
            NewRow := Matrix(K,1,(MapDegree+1)*2,MonomialsNumerator cat MonomialsDenominator);
LinearEquationMatrix:=VerticalJoin(LinearEquationMatrix, NewRow);                // add the corresponding equation to the system
        end if;
    end if;
    end if;
    catch e;
    end try;
end for;


MatrixKernel:=Kernel(Transpose(LinearEquationMatrix));                 // solve the relevant linear system

if Dimension(MatrixKernel) ne 1 then
    print "Failed to determine map with translation vector (", TranslateX, TranslateY, ")";
else
    NumeratorRF := T![(MatrixKernel.1)[i] : i in [1..MapDegree+1]];
    DenominatorRF := T![(MatrixKernel.1)[i] : i in [MapDegree+2..2*(MapDegree+1)]];
    RationalFunction := NumeratorRF/DenominatorRF;                 // this is the map! Or rather, its x-coordinate.

    "C", HyperellipticCurve(f1);
    "K", K;
    "E", WeierstrassModel(E);
    "x-coordinate of the map C to E: ", RationalFunction;

	try	
	        // After clearing the denominator (with a square) and dividing by y^2=f1, the following expression should be a square
		r:=SquareRoot(T!(Evaluate(f2,RationalFunction)*DenominatorRF^4/T!f1));
		"y-coordinate of the map C to E: y *", r/DenominatorRF^2;

		PrintFile("./DecomposableJacobians.out", HyperellipticCurve(f1));
		PrintFile("./DecomposableJacobians.out", K);
		PrintFile("./DecomposableJacobians.out", E);
		PrintFile("./DecomposableJacobians.out", "Degree of map: " cat IntegerToString(MapDegree));
	        PrintFile("./DecomposableJacobians.out", RationalFunction);
	        PrintFile("./DecomposableJacobians.out", r/DenominatorRF^2);

		// Final test: using the fact that y^2=f1, the pair (RationalFunction, y*r/DenominatorRF^2) is a map C \to E iff the following expression evaluates to 0
		return (Evaluate(f2,RationalFunction)-r^2*T!f1/DenominatorRF^4 eq 0);
	catch e
		print e;
	end try;
end if;


end for;
end for;
end for;

catch e;
end try;
end for;
end for;

return false;

end function;


function ComputeMap2(f,h : WhichEigenvalue:=1);
	return ComputeMap(4*f+h^2 : WhichEigenvalue:=WhichEigenvalue);
end function;
